/* 
 * File:   buffer_config.h
 * Author: cstgeo
 *
 * Created on April 24, 2018, 6:49 PM
 */

#ifndef BUFFER_CONFIG_H
#  define	BUFFER_CONFIG_H
#include "config.h"

#define MAX_SYMBOL_BUFFER_SIZE 0xFF
#define MAX_SYMBOL_BUFFER_POINTER_ARRAY_SIZE (MAX_SYMBOL_BUFFER_SIZE/0x2)
#define MAX_NYBBLE_BUFFER_SIZE (0x2 * MAX_SYMBOL_BUFFER_SIZE - 0x1)

typedef unsigned char nyb;
typedef unsigned char sym;

#define ALT_REG_COUNT 5
#define GAM_REG_COUNT 6
#define LIDAR_REG_COUNT 2
//#define GPS_REG_COUNT 0

typedef enum {
	for_COM   = 1,
	from_LORA = 2,
	from_PC   = 3,
	from_LDR  = 4,
	from_GPS  = 5,
	from_BLE  = 6,
	from_ALT  = 7,
	from_GAM  = 8,
	from_PIC  = 9
} src;

int get_register_count(src source) {
	switch (source) {
		case for_COM:		return 0;
		case from_LORA: return 0;
		case from_PC:		return 0;
		case from_LDR:	return LIDAR_REG_COUNT;
		case from_GPS:	return 0;
		case from_BLE:	return 0;
		case from_ALT:	return ALT_REG_COUNT;
		case from_GAM:	return GAM_REG_COUNT;
		case from_PIC:	return 0;
		default:				return-1;
	}
}

bool is_known_source(src source) {
	switch (source) {
		case for_COM:	
		case from_LORA: 
		case from_PC:	
		case from_LDR:	
		case from_GPS:	
		case from_BLE:	
		case from_ALT:	
		case from_GAM:
		case from_PIC: return true;
		default: 		   return false;
	}
}

void print_source(src source) {
	switch (source) {
		case for_COM:		printf("COMMAND"); break;
		case from_LORA: printf("LORA"); break;
		case from_PC:		printf("PC"); break;
		case from_LDR:	printf("LIDAR"); break;
		case from_GPS:	printf("GPS"); break;
		case from_BLE:	printf("BLE"); break;
		case from_ALT:	printf("ALT"); break;
		case from_GAM:	printf("GAM"); break;
		case from_PIC:  printf("PIC32"); break;
		default:				printf("UNKNOWN");
	}
}

typedef struct {
		int symbol_index;
		int first_chunk_pointer_index; // [!] Do not edit this directly. This is the index of the first symbol pointer.
		int last_chunk_pointer_index; // [!] Do not edit this directly. This is the index of the last symbol pointer.
		int nybble_index; // The current index in the hex-encoded version of the information in the buffer. Divide this by 2 to get the index in the string version of the information in the buffer.
		/* The string version of the information in the buffer.
		 * This is the format of the information used to store values equivalent to (or closer to) raw sensor data and/or command data.
		 * Pointers to each chunk (string) of information are stored in the chunk_pointer array.
		 */
		sym symbol[MAX_SYMBOL_BUFFER_SIZE];
		/* [!] Do not edit this directly. This is an array of pointers to each of the symbol chunks (each of the strings) stored in the buffer.
		 */
		sym * chunk_pointer[MAX_SYMBOL_BUFFER_POINTER_ARRAY_SIZE];
		/* The hex-encoded version of the information in the buffer.
		 * This is the format of the information sent to or received from the LoRa module.
		 */
		nyb nybble[MAX_NYBBLE_BUFFER_SIZE];

		bool (* is_empty)(); // Checks to see if the buffer is empty.
		void (* set_empty)(); // Sets the buffer as empty.
		void (* enqueue)(src source, sym * symbol);
		sym * (* dequeue)();
		void (* add_nybble)(nyb nybble); // This should be called each time the LoRa module sends a character to the PIC within the UART interrupt function. (one symbol == two nybbles)
		void (* add_symbol)(src source, sym symbol, bool is_new_chunk); // This is how data is added one symbol at a time from any of the sensors. (one symbol == two nybbles)
		void (* enqueue_lora_nybbles)(nyb * nybble); // This is how the received LoRa data is added to the queue.
} buffer;

#define new_buffer() {0,0,-1,0,{0},{0},{0}}

bool is_empty(buffer * b) {
	return b->last_chunk_pointer_index < b->first_chunk_pointer_index;
}

void set_empty(buffer * b) {
	while (b->last_chunk_pointer_index != -1) {b->chunk_pointer[b->last_chunk_pointer_index--] = 0; }
	b->first_chunk_pointer_index = 0;
	b->nybble[0] = '\0';
	b->symbol[0] = '\0';
	b->nybble_index = 0;
	b->symbol_index = 0;
}

// converts hex digit (0x0-0xF) to symbol value of hex digit ('0' - 'F')
#define x2c(x) ( (x >  9 ) ? x+'7' : x+'0' )

// converts symbol value of hex digit ('0' - 'F') to hex digit (0x0-0xF)
#define c2x(c) ( (c > '9') ? c-'7' : c-'0' )

const char nybble_lookup[] = {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};

void enqueue(buffer * b, src source, sym * s) {
	int count = get_register_count(source);
	if (s[0] || count) { // ensures information is being added
		if (b->last_chunk_pointer_index != -1) { // if the last pointer index is not -1, add the data separator character ('\0') [in both versions, symbol and nybble]
			b->symbol[b->symbol_index++] = '\0';
			b->nybble[b->nybble_index++] = '0';
			b->nybble[b->nybble_index++] = '0';
		}
		// add the source value before adding the chunk pointer
		b->symbol[b->symbol_index++] = (nyb)source;
		b->nybble[b->nybble_index++] = '0';
		b->nybble[b->nybble_index++] = nybble_lookup[(nyb)source];

		b->chunk_pointer[++b->last_chunk_pointer_index] = b->symbol + b->symbol_index; // add a new pointer to the array for the incoming information.
		
		int i = 0;
		while (s[i] || i < count) {

			b->symbol[b->symbol_index] = s[i];
			b->nybble[b->nybble_index] = nybble_lookup[s[i]/0x10];
			b->nybble[b->nybble_index+1] = nybble_lookup[s[i]%0x10];

			i += 1;
			b->symbol_index += 1;
			b->nybble_index += 2;
		}

		b->symbol[b->symbol_index] = '\0';
		b->nybble[b->nybble_index] = '\0';
	} else {
		printf("[!] Error - Nothing was enqueued from source: "); print_source(source); printf("\t(%x)\n\r", source);
	}
}

sym buffer_empty_error_message[] = "[!] Error - The buffer is empty!";
sym * dequeue(buffer * b) {
	if (b->is_empty()) { return buffer_empty_error_message; }
	unsigned char * p = b->chunk_pointer[b->first_chunk_pointer_index];
	b->chunk_pointer[b->first_chunk_pointer_index++] = 0;
	if (b->last_chunk_pointer_index < b->first_chunk_pointer_index) { b->set_empty(); }
	return p;
}

void add_nybble(buffer * b, nyb n) {
	nyb x = n-'0';
	// If this is the first part of the symbol, set the value of the symbol to (x * 0x10), else add x to the symbol value and increment the symbol index.
	if (b->nybble_index%2 == 0) {
		b->symbol[b->symbol_index] = x * 0x10;
	} else {
		b->symbol[b->symbol_index++] += x;
		b->symbol[b->symbol_index] = '\0';
	}
//	printf("\n\t(%x) %c\n", b->symbol[b->nybble_index/2], b->symbol[b->nybble_index/2]);

	b->nybble[b->nybble_index++] = n;
	b->nybble[b->nybble_index] = '\0';
}

void add_symbol(buffer * b, src source, sym s, bool is_new_chunk) {
	if (is_new_chunk) { // If a new chunk, add a new chunk pointer
		if (b->last_chunk_pointer_index >= 0) { // If this is not the first chunk, add the chunk separator character ('\0')
			b->symbol[b->symbol_index++] = '\0';
			b->nybble[b->nybble_index++] = '0';
			b->nybble[b->nybble_index++] = '0';
		}
		// add the source value before adding the chunk pointer
		b->symbol[b->symbol_index++] = (nyb)source;
		b->nybble[b->nybble_index++] = '0';
		b->nybble[b->nybble_index++] = nybble_lookup[(nyb)source];

		b->chunk_pointer[++b->last_chunk_pointer_index] = b->symbol + b->symbol_index;
	}

	b->symbol[b->symbol_index] = s;
	b->nybble[b->nybble_index] = nybble_lookup[s/0x10];
	b->nybble[b->nybble_index+1] = nybble_lookup[s%0x10];

	b->symbol_index += 1;
	b->nybble_index += 2;
}

sym nyb2symMSB(nyb n) { return ( (n > '9') ? n-'7' : n-'0' ) * 0x10; }
sym nyb2symLSB(nyb n) { return ( (n > '9') ? n-'7' : n-'0' ); }
void enqueue_lora_nybbles(buffer * b, nyb * n) {
	int i;
	sym s;
	int known_remaining_registers = 0;
	bool should_get_source = true;
	src source = (src)0;
	for (i=0; n[i] && n[i+1]; i+=2) {
		s = nyb2symMSB(n[i]) + nyb2symLSB(n[i+1]); // The symbol represented by the two nybbles
		if (should_get_source) {
			source = (src)s;
			if (is_known_source(source)) {
				known_remaining_registers = get_register_count(source);
				should_get_source = false;
				i+=2;
				s = nyb2symMSB(n[i]) + nyb2symLSB(n[i+1]); // The symbol represented by the two nybbles
				b->add_symbol(source, s, true); 
			} else { return; }
		} else
		if (known_remaining_registers == 0 && s == '\0') { // if no known registers are left and the symbol is the null character
			should_get_source = true;
		} else { b->add_symbol(source, s, false); }
		known_remaining_registers -= known_remaining_registers > 0 ? 1 : 0;
	}
}

typedef unsigned short ind; /* Index  Type */

#define LORA_RECEIVED_STRING_DATA_OFFSET 10
#define GPS_RECEIVED_STRING_DATA_OFFSET 1
#define BLE_RECEIVED_STRING_DATA_OFFSET 1
#define LORA_MAX_STRING_INDEX MAX_NYBBLE_BUFFER_SIZE + LORA_RECEIVED_STRING_DATA_OFFSET

sym lora_data_to_send_index = 0;
ind lora_received_string_index = 0;
sym gps_received_string_index = 0;
sym ble_received_string_index = 0;

sym lora_data_to_send[MAX_SYMBOL_BUFFER_SIZE] = {0}; /* TODO: decide what type goes in here (nybble or symbol)... */
sym lora_data_to_send_default[] = "01FF";

sym lora_received_string[MAX_NYBBLE_BUFFER_SIZE + LORA_RECEIVED_STRING_DATA_OFFSET] = {0};
unsigned char * received_lora_data;
#define initialize_received_lora_data received_lora_data = lora_received_string + LORA_RECEIVED_STRING_DATA_OFFSET

unsigned char gps_received_string[MAX_SYMBOL_BUFFER_SIZE] = {0};
unsigned char * received_gps_data;
#define initialize_received_gps_data received_gps_data = gps_received_string + GPS_RECEIVED_STRING_DATA_OFFSET

unsigned char ble_received_string[MAX_SYMBOL_BUFFER_SIZE] = {0};
unsigned char * received_ble_data;
#define initialize_received_ble_data received_ble_data = ble_received_string + BLE_RECEIVED_STRING_DATA_OFFSET


#define add_Buffer(n) buffer n##_queue = new_buffer();\
bool n##_queue_is_empty() { return is_empty(&n##_queue); }\
void empty_##n##_queue() { set_empty(&n##_queue); }\
void enqueue_to_##n##_queue(src source, sym * symbol) { enqueue(&n##_queue, source, symbol); }\
sym * dequeue_from_##n##_queue() { return dequeue(&n##_queue); }\
void add_nybble_to_##n##_queue(nyb nybble) { add_nybble(&n##_queue, nybble); }\
void add_symbol_to_##n##_queue(src source, sym symbol, bool is_new_chunk) { add_symbol(&n##_queue, source, symbol, is_new_chunk); }\
void enqueue_lora_nybbles_to_##n##_queue(nyb * nybble) { enqueue_lora_nybbles(&n##_queue, nybble); }\
void configure_##n##_queue() {\
	n##_queue.is_empty = n##_queue_is_empty;\
	n##_queue.set_empty = empty_##n##_queue;\
	n##_queue.enqueue = enqueue_to_##n##_queue;\
	n##_queue.dequeue = dequeue_from_##n##_queue;\
	n##_queue.add_nybble = add_nybble_to_##n##_queue;\
	n##_queue.add_symbol = add_symbol_to_##n##_queue;\
	n##_queue.enqueue_lora_nybbles = enqueue_lora_nybbles_to_##n##_queue;\
	n##_queue.set_empty();\
}

#define Configure_Buffer(n)	configure_##n##_queue();

#define Configure_Buffers(...) \
FOR_EACH(add_Buffer,__VA_ARGS__)\
void configure_buffers() {\
	printConfigurationStarted(Buffers);\
	FOR_EACH(Configure_Buffer,__VA_ARGS__)\
	printConfigurationCompleted(Buffers);\
}

Configure_Buffers(incoming, outgoing, command, data)
		
struct {
	void (* com )(sym * data);
	void (* lora)(sym * data);
	void (* pc  )(sym * data);
	void (* alt )(sym * data);
	void (* ble )(sym * data);
	void (* gam )(sym * data);
	void (* gps )(sym * data);
	void (* ldr )(sym * data);
	void (* pic )(sym * data);
} print;

void print_com_data(sym * data);
void print_lora_data(sym * data);
void print_pc_data(sym * data);
void print_alt_data(sym * data);
void print_ble_data(sym * data);
void print_gam_data(sym * data);
void print_gps_data(sym * data);
void print_ldr_data(sym * data);
void print_pic_data(sym * data);

void configure_data_printing() {
	printConfigurationStarted(Data_Printing);
	print.com = print_com_data;
//	print.lora = print_lora_data;
//	print.pc = print_pc_data;
	print.alt = print_alt_data;
//  print.ble = print_ble_data;
	print.gam = print_gam_data;
//	print.gps = print_gps_data;
	print.ldr = print_ldr_data;
	print.pic = print_pic_data;
	printConfigurationCompleted(Data_Printing);
}

#endif	/* BUFFER_CONFIG_H */

